@@ -2,12 +2,9 @@ use std::sync::{Arc,Mutex};
 use std::rc::{Rc,Weak};
 use std::cell::RefCell;
 use std::net::Shutdown;
-use std::convert::TryFrom;
 use std::os::unix::io::AsRawFd;
 use mio::*;
 use mio::net::*;
-use mio_uds::UnixStream;
-use mio::unix::UnixReady;
 use rusty_ulid::Ulid;
 use std::io::ErrorKind;
 use std::collections::{HashMap, HashSet};
@@ -24,7 +21,6 @@ use openssl::hash::MessageDigest;
 use openssl::nid;
 use openssl::error::ErrorStack;
 use openssl::ssl::SslVersion;
-use mio_extras::timer::{Timer,Timeout};
 
 use sozu_command::scm_socket::ScmSocket;
 use sozu_command::proxy::{Application,CertFingerprint,CertificateAndKey,
@@ -32,24 +28,25 @@ use sozu_command::proxy::{Application,CertFingerprint,CertificateAndKey,
   ProxyResponseStatus,TlsVersion,ProxyEvent,Query,QueryCertificateType,
   QueryAnswer,QueryAnswerCertificate,ProxyResponseData};
 use sozu_command::logging;
-use sozu_command::buffer::fixed::Buffer;
+use sozu_command::ready::Ready;
 
-use protocol::http::{parser::{RequestState,RRequestLine,hostname_and_port}, answers::{DefaultAnswers, CustomAnswers, HttpAnswers}};
+use protocol::http::{parser::{RequestState,RRequestLine,hostname_and_port}, answers::HttpAnswers};
 use pool::Pool;
 use {AppId,Backend,SessionResult,ConnectionError,Protocol,Readiness,SessionMetrics,
   ProxySession,ProxyConfiguration,AcceptError,BackendConnectAction,BackendConnectionStatus,
   CloseResult};
 use backends::BackendMap;
-use server::{Server,ProxyChannel,ListenToken,ListenPortState,SessionToken,
+use server::{Server,ProxyChannel,ListenToken,SessionToken,
   ListenSession, CONN_RETRIES, push_event};
 use socket::server_bind;
 use trie::*;
 use protocol::{ProtocolResult,Http,Pipe,StickySession};
 use protocol::openssl::TlsHandshake;
-use protocol::http::{DefaultAnswerStatus, TimeoutStatus};
+use protocol::http::DefaultAnswerStatus;
 use protocol::proxy_protocol::expect::ExpectProxyProtocol;
 use retry::RetryPolicy;
 use util::UnwrapLog;
+use timer::TimeoutContainer;
 
 #[derive(Debug,Clone,PartialEq,Eq)]
 pub struct TlsApp {
@@ -71,24 +68,26 @@ pub struct Session {
   back_connected:     BackendConnectionStatus,
   protocol:           Option<State>,
   public_address:     SocketAddr,
-  ssl:                Option<Ssl>,
   pool:               Weak<RefCell<Pool>>,
   sticky_name:        String,
   metrics:            SessionMetrics,
   pub app_id:         Option<String>,
-  timeout:            Timeout,
   last_event:         Instant,
   pub listen_token:   Token,
   connection_attempt: u8,
   peer_address:       Option<SocketAddr>,
   answers:            Rc<RefCell<HttpAnswers>>,
+  front_timeout:      TimeoutContainer,
+  frontend_timeout_duration: Duration,
+  backend_timeout_duration: Duration,
 }
 
 impl Session {
   pub fn new(ssl:Ssl, sock: TcpStream, token: Token, pool: Weak<RefCell<Pool>>,
     public_address: SocketAddr, expect_proxy: bool, sticky_name: String,
-    timeout: Timeout, answers: Rc<RefCell<HttpAnswers>>, listen_token: Token,
-    delay: Duration) -> Session {
+    answers: Rc<RefCell<HttpAnswers>>, listen_token: Token,
+    wait_time: Duration, frontend_timeout_duration: Duration,
+    backend_timeout_duration: Duration, request_timeout_duration: Duration) -> Session {
 
     let peer_address = if expect_proxy {
       // Will be defined later once the expect proxy header has been received and parsed
@@ -98,6 +97,8 @@ impl Session {
     };
 
     let request_id = Ulid::generate();
+    let front_timeout = TimeoutContainer::new(request_timeout_duration, token);
+
     let protocol = if expect_proxy {
       trace!("starting in expect proxy state");
       gauge_add!("protocol.proxy.expect", 1);
@@ -107,7 +108,7 @@ impl Session {
       Some(State::Handshake(TlsHandshake::new(ssl, sock, request_id, peer_address.clone())))
     };
 
-    let metrics = SessionMetrics::new(Some(delay));
+    let metrics = SessionMetrics::new(Some(wait_time));
 
     let mut session = Session {
       frontend_token:     token,
@@ -115,19 +116,20 @@ impl Session {
       back_connected:     BackendConnectionStatus::NotConnected,
       protocol,
       public_address,
-      ssl:                None,
       pool,
       sticky_name,
       metrics,
       app_id:             None,
-      timeout,
       last_event:         Instant::now(),
       listen_token,
       connection_attempt: 0,
       peer_address,
       answers,
+      front_timeout,
+      frontend_timeout_duration,
+      backend_timeout_duration,
     };
-    session.front_readiness().interest = UnixReady::from(Ready::readable()) | UnixReady::hup() | UnixReady::error();
+    session.front_readiness().interest = Ready::readable() | Ready::hup() | Ready::error();
 
     session
   }
@@ -152,7 +154,7 @@ impl Session {
     })
   }
 
-  pub fn set_answer(&mut self, answer: DefaultAnswerStatus, buf: Rc<Vec<u8>>)  {
+  pub fn set_answer(&mut self, answer: DefaultAnswerStatus, buf: Option<Rc<Vec<u8>>>)  {
     self.protocol.as_mut().map(|protocol| {
       if let &mut State::Http(ref mut http) = protocol {
         http.set_answer(answer, buf);
@@ -181,7 +183,8 @@ impl Session {
         }
       }
 
-      error!("failed to upgrade from expect");
+      // currently, only happens in expect proxy protocol with AF_UNSPEC address
+      //error!("failed to upgrade from expect");
       self.protocol = Some(State::Expect(expect, ssl));
       false
 
@@ -197,17 +200,19 @@ impl Session {
         ssl.current_cipher().map(|c| incr!(c.name()));
       });
 
+      let backend_timeout_duration = self.backend_timeout_duration;
       let mut http = Http::new(unwrap_msg!(handshake.stream), self.frontend_token.clone(),
         handshake.request_id, pool, self.public_address.clone(), self.peer_address,
-        self.sticky_name.clone(), Protocol::HTTPS);
+        self.sticky_name.clone(), Protocol::HTTPS, self.answers.clone(),
+        self.front_timeout.take(), self.frontend_timeout_duration,
+        backend_timeout_duration);
 
       http.front_readiness = readiness;
-      http.front_readiness.interest = UnixReady::from(Ready::readable()) | UnixReady::hup() | UnixReady::error();
+      http.front_readiness.interest = Ready::readable() | Ready::hup() | Ready::error();
 
       gauge_add!("protocol.tls.handshake", -1);
       gauge_add!("protocol.https", 1);
 
-      self.ssl = handshake.ssl;
       self.protocol = Some(State::Http(http));
       return true;
     } else if let State::Http(http) = protocol {
@@ -250,6 +255,8 @@ impl Session {
 
       pipe.front_readiness.event = http.front_readiness.event;
       pipe.back_readiness.event  = http.back_readiness.event;
+      pipe.front_timeout = Some(http.front_timeout);
+      pipe.back_timeout = Some(http.back_timeout);
       pipe.set_back_token(back_token);
 
       self.protocol = Some(State::WebSocket(pipe));
@@ -298,8 +305,18 @@ impl Session {
 
   fn readable(&mut self)      -> SessionResult {
     let (upgrade, result) = match *unwrap_msg!(self.protocol.as_mut()) {
-      State::Expect(ref mut expect,_)     => expect.readable(&mut self.metrics),
-      State::Handshake(ref mut handshake) => handshake.readable(&mut self.metrics),
+      State::Expect(ref mut expect,_)     => {
+          if !self.front_timeout.reset() {
+              error!("could not reset front timeout (HTTPS OpenSSL upgrading from Expect)");
+          }
+          expect.readable(&mut self.metrics)
+      },
+      State::Handshake(ref mut handshake) => {
+          if !self.front_timeout.reset() {
+              error!("could not reset front timeout (HTTPS OpenSSL upgrading from Handshake)");
+          }
+          handshake.readable(&mut self.metrics)
+      },
       State::Http(ref mut http)           => (ProtocolResult::Continue, http.readable(&mut self.metrics)),
       State::WebSocket(ref mut pipe)      => (ProtocolResult::Continue, pipe.readable(&mut self.metrics)),
     };
@@ -358,6 +375,7 @@ impl Session {
     }
   }
 
+  /*
   fn front_socket(&self) -> Option<&TcpStream> {
     match unwrap_msg!(self.protocol.as_ref()) {
       &State::Expect(ref expect,_)     => Some(expect.front_socket()),
@@ -366,7 +384,18 @@ impl Session {
       &State::WebSocket(ref pipe)      => Some(pipe.front_socket()),
     }
   }
+  */
 
+  fn front_socket_mut(&mut self) -> Option<&mut TcpStream> {
+    match *unwrap_msg!(self.protocol.as_mut()) {
+      State::Expect(ref mut expect,_)     => Some(expect.front_socket_mut()),
+      State::Handshake(ref mut handshake) => handshake.socket_mut(),
+      State::Http(ref mut http)           => Some(http.front_socket_mut()),
+      State::WebSocket(ref mut pipe)      => Some(pipe.front_socket_mut()),
+    }
+  }
+
+  /*
   fn back_socket(&self)  -> Option<&TcpStream> {
     match unwrap_msg!(self.protocol.as_ref()) {
       &State::Expect(_,_)         => None,
@@ -375,6 +404,16 @@ impl Session {
       &State::WebSocket(ref pipe) => pipe.back_socket(),
     }
   }
+  */
+
+  fn back_socket_mut(&mut self)  -> Option<&mut TcpStream> {
+    match *unwrap_msg!(self.protocol.as_mut()) {
+      State::Expect(_,_)             => None,
+      State::Handshake(_)            => None,
+      State::Http(ref mut http)      => http.back_socket_mut(),
+      State::WebSocket(ref mut pipe) => pipe.back_socket_mut(),
+    }
+  }
 
   fn back_token(&self)   -> Option<Token> {
     match unwrap_msg!(self.protocol.as_ref()) {
@@ -386,8 +425,8 @@ impl Session {
   }
 
   fn set_back_socket(&mut self, sock:TcpStream) {
-    let backend_address = self.backend.as_ref().map(|b| b.borrow().address).unwrap();
-    unwrap_msg!(self.http_mut()).set_back_socket(sock, backend_address)
+    let backend = self.backend.clone();
+    unwrap_msg!(self.http_mut()).set_back_socket(sock, backend)
   }
 
   fn set_back_token(&mut self, token: Token) {
@@ -403,6 +442,7 @@ impl Session {
   }
 
   fn set_back_connected(&mut self, connected: BackendConnectionStatus) {
+    let last = self.back_connected.clone();
     self.back_connected = connected;
 
     if connected == BackendConnectionStatus::Connected {
@@ -417,6 +457,12 @@ impl Session {
           push_event(ProxyEvent::BackendUp(backend.backend_id.clone(), backend.address));
         }
 
+        if let BackendConnectionStatus::Connecting(start) = last {
+          backend.set_connection_time(Instant::now() - start);
+        }
+
+        backend.active_requests += 1;
+
         backend.failures = 0;
         backend.retry_policy.succeed();
       });
@@ -476,6 +522,16 @@ impl Session {
   fn reset_connection_attempt(&mut self) {
     self.connection_attempt = 0;
   }
+
+  fn cancel_timeouts(&mut self) {
+      self.front_timeout.cancel();
+
+      match *unwrap_msg!(self.protocol.as_mut()) {
+          State::Http(ref mut http) => http.cancel_timeouts(),
+          State::WebSocket(ref mut pipe) => pipe.cancel_timeouts(),
+          _ => {},
+      }
+  }
 }
 
 impl ProxySession for Session {
@@ -484,13 +540,14 @@ impl ProxySession for Session {
     //println!("TLS closing[{:?}] temp->front: {:?}, temp->back: {:?}", self.frontend_token, *self.temp.front_buf, *self.temp.back_buf);
     self.http_mut().map(|http| http.close());
     self.metrics.service_stop();
-    if let Some(front_socket) = self.front_socket() {
+    self.cancel_timeouts();
+    if let Some(front_socket) = self.front_socket_mut() {
       if let Err(e) = front_socket.shutdown(Shutdown::Both) {
         if e.kind() != ErrorKind::NotConnected {
           error!("error closing front socket({:?}): {:?}", front_socket, e);
         }
       }
-      if let Err(e) = poll.deregister(front_socket) {
+      if let Err(e) = poll.registry().deregister(front_socket) {
         error!("error deregistering front socket({:?}): {:?}", front_socket, e);
       }
     }
@@ -504,10 +561,22 @@ impl ProxySession for Session {
     //FIXME: should we really pass a token here?
     self.close_backend(Token(0), poll);
 
-    if let Some(State::Http(ref http)) = self.protocol {
+    if let Some(State::Http(ref mut http)) = self.protocol {
       //if the state was initial, the connection was already reset
       if http.request != Some(RequestState::Initial) {
         gauge_add!("http.active_requests", -1);
+
+        if let Some(b) = http.backend_data.as_mut() {
+          let mut backend = b.borrow_mut();
+          backend.active_requests = backend.active_requests.saturating_sub(1);
+        }
+      }
+    }
+
+    if let Some(State::WebSocket(_)) = self.protocol {
+      if let Some(b) = self.backend.as_mut() {
+        let mut backend = b.borrow_mut();
+        backend.active_requests = backend.active_requests.saturating_sub(1);
       }
     }
 
@@ -524,37 +593,25 @@ impl ProxySession for Session {
     result
   }
 
-  fn timeout(&mut self, token: Token, timer: &mut Timer<Token>, front_timeout: &Duration) -> SessionResult {
-    if self.frontend_token == token {
-      let dur = Instant::now() - self.last_event;
-      if dur < *front_timeout {
-        timer.set_timeout(std::time::Duration::try_from(*front_timeout - dur).unwrap(), token);
-        SessionResult::Continue
-      } else {
-        match self.http().map(|h| h.timeout_status()) {
-          Some(TimeoutStatus::Request) => {
-            let answer = self.answers.borrow().get(DefaultAnswerStatus::Answer408, None);
-            self.set_answer(DefaultAnswerStatus::Answer408, answer);
-            self.writable()
-          },
-          Some(TimeoutStatus::Response) => {
-            let answer = self.answers.borrow().get(DefaultAnswerStatus::Answer504, None);
-            self.set_answer(DefaultAnswerStatus::Answer504, answer);
-            self.writable()
-          },
-          _ => {
-            SessionResult::CloseSession
+  fn timeout(&mut self, token: Token) -> SessionResult {
+    match *unwrap_msg!(self.protocol.as_mut()) {
+      State::Expect(_,_) => {
+          if token == self.frontend_token {
+              self.front_timeout.triggered();
           }
-        }
-      }
-    } else {
-      // invalid token, obsolete timeout triggered
-      SessionResult::Continue
-    }
-  }
 
-  fn cancel_timeouts(&self, timer: &mut Timer<Token>) {
-    timer.cancel_timeout(&self.timeout);
+          SessionResult::CloseSession
+      },
+      State::Handshake(_) => {
+          if token == self.frontend_token {
+              self.front_timeout.triggered();
+          }
+
+          SessionResult::CloseSession
+      },
+      State::WebSocket(ref mut pipe) => pipe.timeout(token, &mut self.metrics),
+      State::Http(ref mut http) => http.timeout(token, &mut self.metrics),
+    }
   }
 
   //FIXME: check the token passed as argument
@@ -563,14 +620,14 @@ impl ProxySession for Session {
 
     let back_connected = self.back_connected();
     if back_connected != BackendConnectionStatus::NotConnected {
-       self.back_readiness().map(|r| r.event = UnixReady::from(Ready::empty()));
-      if let Some(sock) = self.back_socket() {
+       self.back_readiness().map(|r| r.event = Ready::empty());
+      if let Some(sock) = self.back_socket_mut() {
         if let Err(e) = sock.shutdown(Shutdown::Both) {
           if e.kind() != ErrorKind::NotConnected {
             error!("error closing back socket({:?}): {:?}", sock, e);
           }
         }
-        if let Err(e) = poll.deregister(sock) {
+        if let Err(e) = poll.registry().deregister(sock) {
           error!("error deregistering back socket({:?}): {:?}", sock, e);
         }
       }
@@ -593,14 +650,14 @@ impl ProxySession for Session {
   }
 
   fn process_events(&mut self, token: Token, events: Ready) {
-    trace!("token {:?} got event {}", token, super::unix_ready_to_string(UnixReady::from(events)));
+    trace!("token {:?} got event {}", token, super::ready_to_string(events));
     self.last_event = Instant::now();
     self.metrics.wait_start();
 
     if self.frontend_token == token {
-      self.front_readiness().event = self.front_readiness().event | UnixReady::from(events);
+      self.front_readiness().event = self.front_readiness().event | events;
     } else if self.back_token() == Some(token) {
-      self.back_readiness().map(|r| r.event = r.event | UnixReady::from(events));
+      self.back_readiness().map(|r| r.event = r.event | events);
     }
   }
 
@@ -610,8 +667,10 @@ impl ProxySession for Session {
 
     self.metrics().service_start();
 
-    if self.back_connected() == BackendConnectionStatus::Connecting &&
-      self.back_readiness().map(|r| r.event != UnixReady::from(Ready::empty())).unwrap_or(false) {
+    if self.back_connected().is_connecting() &&
+      self.back_readiness().map(|r| r.event != Ready::empty()).unwrap_or(false) {
+
+      self.http_mut().map(|h| h.cancel_backend_timeout());
 
       if self.back_readiness().map(|r| r.event.is_hup()).unwrap_or(false) ||
         !self.http_mut().map(|h| h.test_back_socket()).unwrap_or(false) {
@@ -623,6 +682,7 @@ impl ProxySession for Session {
         self.fail_backend_connection();
 
         let backend_token = self.back_token();
+        self.back_connected = BackendConnectionStatus::Connecting(Instant::now());
         return SessionResult::ReconnectBackend(Some(self.frontend_token), backend_token);
       } else {
         self.metrics().backend_connected();
@@ -638,7 +698,7 @@ impl ProxySession for Session {
           return order;
         },
         _ => {
-          self.front_readiness().event.remove(UnixReady::hup());
+          self.front_readiness().event.remove(Ready::hup());
           return order;
         }
       }
@@ -647,11 +707,11 @@ impl ProxySession for Session {
     let token = self.frontend_token.clone();
     while counter < max_loop_iterations {
       let front_interest = self.front_readiness().interest & self.front_readiness().event;
-      let back_interest  = self.back_readiness().map(|r| r.interest & r.event).unwrap_or(UnixReady::from(Ready::empty()));
+      let back_interest  = self.back_readiness().map(|r| r.interest & r.event).unwrap_or(Ready::empty());
 
       trace!("PROXY\t{} {:?} {:?} -> {:?}", self.log_context(), token, self.front_readiness().clone(), self.back_readiness());
 
-      if front_interest == UnixReady::from(Ready::empty()) && back_interest == UnixReady::from(Ready::empty()) {
+      if front_interest == Ready::empty() && back_interest == Ready::empty() {
         break;
       }
 
@@ -705,7 +765,7 @@ impl ProxySession for Session {
             }*/
           },
           _ => {
-            self.back_readiness().map(|r| r.event.remove(UnixReady::hup()));
+            self.back_readiness().map(|r| r.event.remove(Ready::hup()));
             return order;
           }
         };
@@ -713,16 +773,15 @@ impl ProxySession for Session {
 
       if front_interest.is_error() {
           error!("PROXY session {:?} front error, disconnecting", self.frontend_token);
-
-          self.front_readiness().interest = UnixReady::from(Ready::empty());
-          self.back_readiness().map(|r| r.interest  = UnixReady::from(Ready::empty()));
-              return SessionResult::CloseSession;
+          self.front_readiness().interest = Ready::empty();
+          self.back_readiness().map(|r| r.interest  = Ready::empty());
+          return SessionResult::CloseSession;
       }
 
       if back_interest.is_error() {
           if self.back_hup() == SessionResult::CloseSession {
-              self.front_readiness().interest = UnixReady::from(Ready::empty());
-              self.back_readiness().map(|r| r.interest  = UnixReady::from(Ready::empty()));
+              self.front_readiness().interest = Ready::empty();
+              self.back_readiness().map(|r| r.interest  = Ready::empty());
               error!("PROXY session {:?} back error, disconnecting", self.frontend_token);
               return SessionResult::CloseSession;
           }
@@ -837,7 +896,7 @@ pub struct Listener {
   default_context: SslContext,
   answers:         Rc<RefCell<HttpAnswers>>,
   config:          HttpsListener,
-  ssl_options:     SslOptions,
+  _ssl_options:    SslOptions,
   pub token:       Token,
   active:          bool,
 }
@@ -860,13 +919,13 @@ impl Listener {
       listener:        None,
       address:         config.front.clone(),
       domains:         rc_domains,
-      default_context: default_context,
+      default_context,
       contexts:        rc_ctx,
       answers:         Rc::new(RefCell::new(HttpAnswers::new(&config.answer_404, &config.answer_503))),
       active:          false,
       fronts,
       config,
-      ssl_options,
+      _ssl_options: ssl_options,
       token,
     }
   }
@@ -876,13 +935,13 @@ impl Listener {
       return Some(self.token);
     }
 
-    let listener = tcp_listener.or_else(|| server_bind(&self.config.front).map_err(|e| {
+    let mut listener = tcp_listener.or_else(|| server_bind(&self.config.front).map_err(|e| {
       error!("could not create listener {:?}: {:?}", self.config.front, e);
     }).ok());
 
 
-    if let Some(ref sock) = listener {
-      if let Err(e) = event_loop.register(sock, self.token, Ready::readable(), PollOpt::edge()) {
+    if let Some(ref mut sock) = listener {
+      if let Err(e) = event_loop.registry().register(sock, self.token, Interest::READABLE) {
         error!("cannot register listen socket({:?}): {:?}", sock, e);
       }
     } else {
@@ -950,7 +1009,7 @@ impl Listener {
         TlsVersion::TLSv1_2 => versions.remove(ssl::SslOptions::NO_TLSV1_2),
         //TlsVersion::TLSv1_3 => versions.remove(ssl::SslOptions::NO_TLSV1_3),
         TlsVersion::TLSv1_3 => {},
-        s         => error!("unrecognized TLS version: {:?}", s)
+        //s         => error!("unrecognized TLS version: {:?}", s)
       };
     }
 
@@ -992,6 +1051,7 @@ impl Listener {
     Some((context.build(), ssl_options))
   }
 
+  #[allow(dead_code)]
   fn create_servername_callback(ref_ctx: Arc<Mutex<HashMap<CertFingerprint,TlsData>>>,
     ref_domains: Arc<Mutex<TrieNode<CertFingerprint>>>)
     -> impl Fn(&mut SslRef, &mut SslAlert) -> Result<(), SniError> + 'static + Sync + Send {
@@ -1120,7 +1180,7 @@ impl Listener {
           &f.hostname == &front.hostname &&
           &f.path_begin == &front.path_begin
         }) {
-          let front = fronts.remove(pos);
+          let _front = fronts.remove(pos);
         }
       }
 
@@ -1254,7 +1314,7 @@ impl Listener {
     }
   }
 
-  fn accept(&mut self, token: ListenToken) -> Result<TcpStream, AcceptError> {
+  fn accept(&mut self, _token: ListenToken) -> Result<TcpStream, AcceptError> {
     if let Some(ref sock) = self.listener {
       sock.accept().map_err(|e| {
         match e.kind() {
@@ -1368,8 +1428,7 @@ impl Proxy {
 
     match res {
       Err(e) => {
-        let answer = self.get_service_unavailable_answer(Some(app_id), &session.listen_token);
-        session.set_answer(DefaultAnswerStatus::Answer503, answer);
+        session.set_answer(DefaultAnswerStatus::Answer503, None);
         Err(e)
       },
       Ok((backend, conn))  => {
@@ -1400,8 +1459,7 @@ impl Proxy {
     let host: &str = if let Ok((i, (hostname, port))) = hostname_and_port(h.as_bytes()) {
       if i != &b""[..] {
         error!("connect_to_backend: invalid remaining chars after hostname. Host: {}", h);
-        let answer = self.listeners[&session.listen_token].answers.borrow().get(DefaultAnswerStatus::Answer400, None);
-        session.set_answer(DefaultAnswerStatus::Answer400, answer);
+        session.set_answer(DefaultAnswerStatus::Answer400, None);
         return Err(ConnectionError::InvalidHost);
       }
 
@@ -1417,8 +1475,7 @@ impl Proxy {
       if servername.as_ref().map(|s| s.as_str()) != Some(hostname_str) {
         error!("TLS SNI hostname '{:?}' and Host header '{}' don't match", servername, hostname_str);
         /*FIXME: deactivate this check for a temporary test
-        let answer = self.listeners[&session.listen_token].answers.borrow().get(DefaultAnswerStatus::Answer404, None);
-        unwrap_msg!(session.http()).set_answer(DefaultAnswerStatus::Answer404, answer);
+        unwrap_msg!(session.http()).set_answer(DefaultAnswerStatus::Answer404, None);
         return Err(ConnectionError::HostNotFound);
         */
       }
@@ -1432,8 +1489,7 @@ impl Proxy {
       }
     } else {
       error!("hostname parsing failed");
-      let answer = self.listeners[&session.listen_token].answers.borrow().get(DefaultAnswerStatus::Answer400, None);
-      session.set_answer(DefaultAnswerStatus::Answer400, answer);
+      session.set_answer(DefaultAnswerStatus::Answer400, None);
       return Err(ConnectionError::InvalidHost);
     };
 
@@ -1444,8 +1500,7 @@ impl Proxy {
       .map(|ref front| front.app_id.clone()) {
       Some(app_id) => Ok(app_id),
       None => {
-        let answer = self.listeners[&session.listen_token].answers.borrow().get(DefaultAnswerStatus::Answer404, None);
-        session.set_answer(DefaultAnswerStatus::Answer404, answer);
+        session.set_answer(DefaultAnswerStatus::Answer404, None);
         Err(ConnectionError::HostNotFound)
       }
     }
@@ -1454,17 +1509,12 @@ impl Proxy {
   fn check_circuit_breaker(&mut self, session: &mut Session) -> Result<(), ConnectionError> {
     if session.connection_attempt == CONN_RETRIES {
       error!("{} max connection attempt reached", session.log_context());
-      let answer = self.get_service_unavailable_answer(session.app_id.as_ref().map(|app_id| app_id.as_str()), &session.listen_token);
-      session.set_answer(DefaultAnswerStatus::Answer503, answer);
+      session.set_answer(DefaultAnswerStatus::Answer503, None);
       Err(ConnectionError::NoBackendAvailable)
     } else {
       Ok(())
     }
   }
-
-  fn get_service_unavailable_answer(&self, app_id: Option<&str>, listen_token: &Token) -> Rc<Vec<u8>> {
-    self.listeners[&listen_token].answers.borrow().get(DefaultAnswerStatus::Answer503, app_id)
-  }
 }
 
 impl ProxyConfiguration<Session> for Proxy {
@@ -1472,18 +1522,18 @@ impl ProxyConfiguration<Session> for Proxy {
     self.listeners.get_mut(&Token(token.0)).unwrap().accept(token)
   }
 
-  fn create_session(&mut self, frontend_sock: TcpStream, token: ListenToken, poll: &mut Poll, session_token: Token, timeout: Timeout, delay: Duration)
+  fn create_session(&mut self, mut frontend_sock: TcpStream, token: ListenToken, poll: &mut Poll, session_token: Token,
+                    wait_time: Duration)
     -> Result<(Rc<RefCell<Session>>, bool), AcceptError> {
     if let Some(ref listener) = self.listeners.get(&Token(token.0)) {
       if let Err(e) = frontend_sock.set_nodelay(true) {
         error!("error setting nodelay on front socket({:?}): {:?}", frontend_sock, e);
       }
       if let Ok(ssl) = Ssl::new(&listener.default_context) {
-        if let Err(e) = poll.register(
-          &frontend_sock,
+        if let Err(e) = poll.registry().register(
+          &mut frontend_sock,
           session_token,
-          Ready::readable() | Ready::writable() | Ready::from(UnixReady::hup() | UnixReady::error()),
-          PollOpt::edge()
+          Interest::READABLE | Interest::WRITABLE
           ) {
           error!("error registering front socket({:?}): {:?}", frontend_sock, e);
         }
@@ -1491,7 +1541,11 @@ impl ProxyConfiguration<Session> for Proxy {
         let c = Session::new(ssl, frontend_sock, session_token, Rc::downgrade(&self.pool),
           listener.config.public_address.unwrap_or(listener.config.front),
           listener.config.expect_proxy, listener.config.sticky_name.clone(),
-          timeout, listener.answers.clone(), Token(token.0), delay);
+          listener.answers.clone(), Token(token.0), wait_time,
+          Duration::seconds(listener.config.front_timeout as i64),
+          Duration::seconds(listener.config.back_timeout as i64),
+          Duration::seconds(listener.config.request_timeout as i64),
+          );
 
         Ok((Rc::new(RefCell::new(c)), false))
       } else {
@@ -1524,6 +1578,7 @@ impl ProxyConfiguration<Session> for Proxy {
         //matched on keepalive
         session.metrics.backend_id = session.backend.as_ref().map(|i| i.borrow().backend_id.clone());
         session.metrics.backend_start();
+        session.http_mut().map(|h| h.backend_data.as_mut().map(|b| b.borrow_mut().active_requests += 1));
         return Ok(BackendConnectAction::Reuse);
       } else {
         if let Some(token) = session.back_token() {
@@ -1550,7 +1605,7 @@ impl ProxyConfiguration<Session> for Proxy {
     session.app_id = Some(app_id.clone());
 
     let front_should_stick = self.applications.get(&app_id).map(|ref app| app.sticky_session).unwrap_or(false);
-    let socket = self.backend_from_request(session, &app_id, front_should_stick)?;
+    let mut socket = self.backend_from_request(session, &app_id, front_should_stick)?;
 
     session.http_mut().map(|http| {
       http.app_id = Some(app_id.clone());
@@ -1560,36 +1615,37 @@ impl ProxyConfiguration<Session> for Proxy {
       error!("error setting nodelay on back socket({:?}): {:?}", socket, e);
     }
     session.back_readiness().map(|r| {
-      r.interest = UnixReady::from(Ready::writable()) | UnixReady::hup() | UnixReady::error();
+      r.interest = Ready::writable() | Ready::hup() | Ready::error();
     });
 
+    let connect_timeout = time::Duration::seconds(i64::from(self.listeners.get(&session.listen_token).as_ref().map(|l| l.config.connect_timeout).unwrap()));
 
-    session.back_connected = BackendConnectionStatus::Connecting;
+    session.back_connected = BackendConnectionStatus::Connecting(Instant::now());
     if let Some(back_token) = old_back_token {
       session.set_back_token(back_token);
-      if let Err(e) = poll.register(
-        &socket,
+      if let Err(e) = poll.registry().register(
+        &mut socket,
         back_token,
-        Ready::readable() | Ready::writable() | Ready::from(UnixReady::hup() | UnixReady::error()),
-        PollOpt::edge()
+        Interest::READABLE | Interest::WRITABLE
       ) {
         error!("error registering back socket({:?}): {:?}", socket, e);
       }
 
       session.set_back_socket(socket);
+      session.http_mut().map(|h| h.set_back_timeout(connect_timeout));
       Ok(BackendConnectAction::Replace)
     } else {
-      if let Err(e) = poll.register(
-        &socket,
+      if let Err(e) = poll.registry().register(
+        &mut socket,
         back_token,
-        Ready::readable() | Ready::writable() | Ready::from(UnixReady::hup() | UnixReady::error()),
-        PollOpt::edge()
+        Interest::READABLE | Interest::WRITABLE
       ) {
         error!("error registering back socket({:?}): {:?}", socket, e);
       }
 
       session.set_back_socket(socket);
       session.set_back_token(back_token);
+      session.http_mut().map(|h| h.set_back_timeout(connect_timeout));
       Ok(BackendConnectAction::New)
     }
   }
@@ -1672,8 +1728,8 @@ impl ProxyConfiguration<Session> for Proxy {
       ProxyRequestData::SoftStop => {
         info!("{} processing soft shutdown", message.id);
         for (_, l) in self.listeners.iter_mut() {
-          l.listener.take().map(|sock| {
-            if let Err(e) = event_loop.deregister(&sock) {
+          l.listener.take().map(|mut sock| {
+            if let Err(e) = event_loop.registry().deregister(&mut sock) {
               error!("error deregistering listen socket({:?}): {:?}", sock, e);
             }
           });
@@ -1683,8 +1739,8 @@ impl ProxyConfiguration<Session> for Proxy {
       ProxyRequestData::HardStop => {
         info!("{} hard shutdown", message.id);
         for (_, mut l) in self.listeners.drain(){
-          l.listener.take().map(|sock| {
-            if let Err(e) = event_loop.deregister(&sock) {
+          l.listener.take().map(|mut sock| {
+            if let Err(e) = event_loop.registry().deregister(&mut sock) {
               error!("error dereginstering listen socket({:?}): {:?}", sock, e);
             }
           });
@@ -1704,7 +1760,7 @@ impl ProxyConfiguration<Session> for Proxy {
         ProxyResponse{ id: message.id, status: ProxyResponseStatus::Ok, data: None }
       },
       ProxyRequestData::Query(Query::Certificates(QueryCertificateType::All)) => {
-        let res = self.listeners.iter().map(|(addr, listener)| {
+        let res = self.listeners.values().map(|listener| {
           let mut domains = unwrap_msg!(listener.domains.lock()).to_hashmap();
           let res = domains.drain().map(|(k, v)| {
             (String::from_utf8(k).unwrap(), v.0.clone())
@@ -1719,7 +1775,7 @@ impl ProxyConfiguration<Session> for Proxy {
           data: Some(ProxyResponseData::Query(QueryAnswer::Certificates(QueryAnswerCertificate::All(res)))) }
       },
       ProxyRequestData::Query(Query::Certificates(QueryCertificateType::Domain(d))) => {
-        let res = self.listeners.iter().map(|(addr, listener)| {
+        let res = self.listeners.values().map(|listener| {
           let domains  = unwrap_msg!(listener.domains.lock());
           (listener.address, domains.domain_lookup(d.as_bytes(), true).map(|(k, v)| {
             (String::from_utf8(k.to_vec()).unwrap(), v.0.clone())
@@ -1737,12 +1793,6 @@ impl ProxyConfiguration<Session> for Proxy {
       }
     }
   }
-
-  fn listen_port_state(&self, port: &u16) -> ListenPortState {
-    fixme!();
-    ListenPortState::Available
-    //if port == &self.address.port() { ListenPortState::InUse } else { ListenPortState::Available }
-  }
 }
 
 
@@ -1803,31 +1853,32 @@ pub fn start(config: HttpsListener, channel: ProxyChannel, max_buffers: usize, b
   let mut event_loop  = Poll::new().expect("could not create event loop");
 
   let pool = Rc::new(RefCell::new(
-    Pool::with_capacity(2*max_buffers, buffer_size)
+    Pool::with_capacity(1, max_buffers, buffer_size)
   ));
   let backends = Rc::new(RefCell::new(BackendMap::new()));
 
-  let mut sessions: Slab<Rc<RefCell<ProxySessionCast>>,SessionToken> = Slab::with_capacity(max_buffers);
+  let mut sessions: Slab<Rc<RefCell<dyn ProxySessionCast>>> = Slab::with_capacity(max_buffers);
   {
-    let entry = sessions.vacant_entry().expect("session list should have enough room at startup");
-    info!("taking token {:?} for channel", entry.index());
+    let entry = sessions.vacant_entry();
+    info!("taking token {:?} for channel", SessionToken(entry.key()));
     entry.insert(Rc::new(RefCell::new(ListenSession { protocol: Protocol::HTTPListen })));
   }
   {
-    let entry = sessions.vacant_entry().expect("session list should have enough room at startup");
-    info!("taking token {:?} for timer", entry.index());
+    let entry = sessions.vacant_entry();
+    info!("taking token {:?} for timer", SessionToken(entry.key()));
     entry.insert(Rc::new(RefCell::new(ListenSession { protocol: Protocol::HTTPListen })));
   }
   {
-    let entry = sessions.vacant_entry().expect("session list should have enough room at startup");
-    info!("taking token {:?} for metrics", entry.index());
+    let entry = sessions.vacant_entry();
+    info!("taking token {:?} for metrics", SessionToken(entry.key()));
     entry.insert(Rc::new(RefCell::new(ListenSession { protocol: Protocol::HTTPListen })));
   }
 
   let token = {
-    let entry = sessions.vacant_entry().expect("session list should have enough room at startup");
-    let e = entry.insert(Rc::new(RefCell::new(ListenSession { protocol: Protocol::HTTPListen })));
-    Token(e.index().0)
+    let entry = sessions.vacant_entry();
+    let key = entry.key();
+    let _e = entry.insert(Rc::new(RefCell::new(ListenSession { protocol: Protocol::HTTPListen })));
+    Token(key)
   };
 
   let mut configuration = Proxy::new(pool.clone(), backends.clone());
@@ -1867,12 +1918,12 @@ mod tests {
   #[cfg(target_pointer_width = "64")]
   fn size_test() {
     assert_size!(ExpectProxyProtocol<mio::net::TcpStream>, 520);
-    assert_size!(TlsHandshake, 216);
-    assert_size!(Http<SslStream<mio::net::TcpStream>>, 1016);
-    assert_size!(Pipe<SslStream<mio::net::TcpStream>>, 224);
-    assert_size!(State, 1024);
+    assert_size!(TlsHandshake, 240);
+    assert_size!(Http<SslStream<mio::net::TcpStream>>, 1232);
+    assert_size!(Pipe<SslStream<mio::net::TcpStream>>, 272);
+    assert_size!(State, 1240);
     // fails depending on the platform?
-    //assert_size!(Session, 1320);
+    assert_size!(Session, 1672);
 
     assert_size!(SslStream<mio::net::TcpStream>, 16);
     assert_size!(Ssl, 8);
@@ -1923,7 +1974,7 @@ mod tests {
       contexts: rc_ctx,
       answers:   Rc::new(RefCell::new(HttpAnswers::new("HTTP/1.1 404 Not Found\r\n\r\n", "HTTP/1.1 503 your application is in deployment\r\n\r\n"))),
       config: Default::default(),
-      ssl_options: ssl::SslOptions::CIPHER_SERVER_PREFERENCE | ssl::SslOptions::NO_COMPRESSION | ssl::SslOptions::NO_TICKET |
+      _ssl_options: ssl::SslOptions::CIPHER_SERVER_PREFERENCE | ssl::SslOptions::NO_COMPRESSION | ssl::SslOptions::NO_TICKET |
         ssl::SslOptions::NO_SSLV2 | ssl::SslOptions::NO_SSLV3 | ssl::SslOptions::NO_TLSV1 | ssl::SslOptions::NO_TLSV1_1,
       token: Token(0),
       active: true,
@@ -2084,7 +2135,7 @@ fn parse_sni_name_list(i: &[u8]) -> Option<&[u8]> {
 fn parse_sni_name(i: &[u8]) -> Option<(&[u8], &[u8])> {
     use nom::{multi::length_data, sequence::preceded,
       bytes::complete::tag,
-      number::complete::{be_u8, be_u16}
+      number::complete::be_u16
     };
 
     if i.is_empty() {
