@@ -1,16 +1,12 @@
 use std::rc::{Rc,Weak};
 use std::cell::RefCell;
-use std::convert::TryFrom;
 use std::net::{Shutdown,SocketAddr};
 use mio::*;
 use mio::net::*;
-use mio::unix::UnixReady;
-use std::io::{ErrorKind,Read};
+use std::io::{ErrorKind, Read};
 use time::{Instant, Duration};
 use rusty_ulid::Ulid;
 use rustls::{ServerSession,Session as ClientSession,ProtocolVersion,SupportedCipherSuite,CipherSuite};
-use mio_extras::timer::{Timer, Timeout};
-use sozu_command::buffer::fixed::Buffer;
 use sozu_command::proxy::ProxyEvent;
 
 use protocol::http::parser::RequestState;
@@ -20,12 +16,14 @@ use {Backend,SessionResult,Protocol,Readiness,SessionMetrics, ProxySession,
 use socket::FrontRustls;
 use protocol::{ProtocolResult,Http,Pipe};
 use protocol::rustls::TlsHandshake;
-use protocol::http::{DefaultAnswerStatus, TimeoutStatus, answers::HttpAnswers};
+use protocol::http::{DefaultAnswerStatus, answers::HttpAnswers};
 use protocol::proxy_protocol::expect::ExpectProxyProtocol;
 use retry::RetryPolicy;
 use util::UnwrapLog;
 use buffer_queue::BufferQueue;
 use server::push_event;
+use timer::TimeoutContainer;
+use sozu_command::ready::Ready;
 
 pub enum State {
   Expect(ExpectProxyProtocol<TcpStream>, ServerSession),
@@ -44,18 +42,22 @@ pub struct Session {
   pub metrics:        SessionMetrics,
   pub app_id:         Option<String>,
   sticky_name:        String,
-  timeout:            Timeout,
   last_event:         Instant,
   pub listen_token:   Token,
   pub connection_attempt: u8,
   peer_address:       Option<SocketAddr>,
   answers:            Rc<RefCell<HttpAnswers>>,
+  front_timeout:      TimeoutContainer,
+  frontend_timeout_duration: Duration,
+  backend_timeout_duration: Duration,
 }
 
 impl Session {
   pub fn new(ssl: ServerSession, sock: TcpStream, token: Token, pool: Weak<RefCell<Pool>>,
-    public_address: SocketAddr, expect_proxy: bool, sticky_name: String, timeout: Timeout,
-    answers: Rc<RefCell<HttpAnswers>>, listen_token: Token, delay: Duration) -> Session {
+    public_address: SocketAddr, expect_proxy: bool, sticky_name: String,
+    answers: Rc<RefCell<HttpAnswers>>, listen_token: Token, wait_time: Duration,
+    frontend_timeout_duration: Duration, backend_timeout_duration: Duration,
+    request_timeout_duration: Duration) -> Session {
     let peer_address = if expect_proxy {
       // Will be defined later once the expect proxy header has been received and parsed
       None
@@ -64,6 +66,8 @@ impl Session {
     };
 
     let request_id = Ulid::generate();
+    let front_timeout = TimeoutContainer::new(request_timeout_duration, token);
+
     let state = if expect_proxy {
       trace!("starting in expect proxy state");
       gauge_add!("protocol.proxy.expect", 1);
@@ -73,7 +77,7 @@ impl Session {
       Some(State::Handshake(TlsHandshake::new(ssl, sock, request_id)))
     };
 
-    let metrics = SessionMetrics::new(Some(delay));
+    let metrics = SessionMetrics::new(Some(wait_time));
 
     let mut session = Session {
       frontend_token: token,
@@ -85,14 +89,16 @@ impl Session {
       metrics,
       app_id:         None,
       sticky_name,
-      timeout,
       last_event:     Instant::now(),
       listen_token,
       connection_attempt: 0,
       peer_address,
       answers,
+      front_timeout,
+      frontend_timeout_duration,
+      backend_timeout_duration,
     };
-    session.front_readiness().interest = UnixReady::from(Ready::readable()) | UnixReady::hup() | UnixReady::error();
+    session.front_readiness().interest = Ready::readable() | Ready::hup() | Ready::error();
     session
   }
 
@@ -116,7 +122,7 @@ impl Session {
     })
   }
 
-  pub fn set_answer(&mut self, answer: DefaultAnswerStatus, buf: Rc<Vec<u8>>)  {
+  pub fn set_answer(&mut self, answer: DefaultAnswerStatus, buf: Option<Rc<Vec<u8>>>)  {
     self.protocol.as_mut().map(|protocol| {
       if let State::Http(ref mut http) = *protocol {
         http.set_answer(answer, buf);
@@ -148,7 +154,8 @@ impl Session {
         }
       }
 
-      error!("failed to upgrade from expect");
+      // currently, only happens in expect proxy protocol with AF_UNSPEC address
+      //error!("failed to upgrade from expect");
       self.protocol = Some(State::Expect(expect, ssl));
       false
     } else if let State::Handshake(handshake) = protocol {
@@ -175,7 +182,9 @@ impl Session {
       let readiness = handshake.readiness.clone();
       let mut http = Http::new(front_stream, self.frontend_token, handshake.request_id,
                                self.pool.clone(), self.public_address, self.peer_address,
-                               self.sticky_name.clone(), Protocol::HTTPS);
+                               self.sticky_name.clone(), Protocol::HTTPS, self.answers.clone(),
+                               self.front_timeout.take(),
+                               self.frontend_timeout_duration, self.backend_timeout_duration);
 
       let res = http.frontend.session.read(front_buf.space());
       match res {
@@ -198,7 +207,7 @@ impl Session {
       gauge_add!("protocol.https", 1);
       http.front_buf = Some(buf);
       http.front_readiness = readiness;
-      http.front_readiness.interest = UnixReady::from(Ready::readable()) | UnixReady::hup() | UnixReady::error();
+      http.front_readiness.interest = Ready::readable() | Ready::hup() | Ready::error();
 
       self.protocol = Some(State::Http(http));
       return true;
@@ -238,6 +247,8 @@ impl Session {
 
       pipe.front_readiness.event = http.front_readiness.event;
       pipe.back_readiness.event  = http.back_readiness.event;
+      pipe.front_timeout = Some(http.front_timeout);
+      pipe.back_timeout = Some(http.back_timeout);
       pipe.set_back_token(back_token);
       pipe.set_app_id(self.app_id.clone());
 
@@ -290,8 +301,18 @@ impl Session {
 
   fn readable(&mut self)      -> SessionResult {
     let (upgrade, result) = match *unwrap_msg!(self.protocol.as_mut()) {
-      State::Expect(ref mut expect, _)    => expect.readable(&mut self.metrics),
-      State::Handshake(ref mut handshake) => handshake.readable(),
+      State::Expect(ref mut expect, _)    => {
+          if !self.front_timeout.reset() {
+              error!("could not reset front timeout");
+          }
+          expect.readable(&mut self.metrics)
+      },
+      State::Handshake(ref mut handshake) => {
+          if !self.front_timeout.reset() {
+              error!("could not reset front timeout");
+          }
+          handshake.readable()
+      },
       State::Http(ref mut http)           => (ProtocolResult::Continue, http.readable(&mut self.metrics)),
       State::WebSocket(ref mut pipe)      => (ProtocolResult::Continue, pipe.readable(&mut self.metrics)),
     };
@@ -364,6 +385,15 @@ impl Session {
     }
   }
 
+  pub fn front_socket_mut(&mut self) -> &mut TcpStream {
+    match unwrap_msg!(self.protocol.as_mut()) {
+      State::Expect(ref mut expect,_)     => expect.front_socket_mut(),
+      State::Handshake(ref mut handshake) => &mut handshake.stream,
+      State::Http(ref mut http)           => http.front_socket_mut(),
+      State::WebSocket(ref mut pipe)      => pipe.front_socket_mut(),
+    }
+  }
+
   pub fn back_socket(&self)  -> Option<&TcpStream> {
     match unwrap_msg!(self.protocol.as_ref()) {
       State::Expect(_,_)         => None,
@@ -373,6 +403,15 @@ impl Session {
     }
   }
 
+  pub fn back_socket_mut(&mut self)  -> Option<&mut TcpStream> {
+    match unwrap_msg!(self.protocol.as_mut()) {
+      State::Expect(_,_)             => None,
+      State::Handshake(_)            => None,
+      State::Http(ref mut http)      => http.back_socket_mut(),
+      State::WebSocket(ref mut pipe) => pipe.back_socket_mut(),
+    }
+  }
+
   pub fn back_token(&self)   -> Option<Token> {
     match unwrap_msg!(self.protocol.as_ref()) {
       State::Expect(_,_)         => None,
@@ -384,7 +423,7 @@ impl Session {
 
   pub fn set_back_socket(&mut self, sock:TcpStream) {
     if let State::Http(ref mut http) = unwrap_msg!(self.protocol.as_mut()) {
-      http.set_back_socket(sock, self.backend.as_ref().map(|b| b.borrow().address).unwrap())
+      http.set_back_socket(sock, self.backend.clone())
     }
   }
 
@@ -401,6 +440,7 @@ impl Session {
   }
 
   fn set_back_connected(&mut self, connected: BackendConnectionStatus) {
+    let last = self.back_connected.clone();
     self.back_connected = connected;
 
     if connected == BackendConnectionStatus::Connected {
@@ -414,6 +454,12 @@ impl Session {
           push_event(ProxyEvent::BackendUp(backend.backend_id.clone(), backend.address));
         }
 
+        if let BackendConnectionStatus::Connecting(start) = last {
+          backend.set_connection_time(Instant::now() - start);
+        }
+
+        backend.active_requests += 1;
+
         backend.failures = 0;
         backend.retry_policy.succeed();
       });
@@ -469,6 +515,16 @@ impl Session {
   fn reset_connection_attempt(&mut self) {
     self.connection_attempt = 0;
   }
+
+  fn cancel_timeouts(&mut self) {
+      self.front_timeout.cancel();
+
+      match *unwrap_msg!(self.protocol.as_mut()) {
+          State::Http(ref mut http) => http.cancel_timeouts(),
+          State::WebSocket(ref mut pipe) => pipe.cancel_timeouts(),
+          _ => {},
+      }
+  }
 }
 
 impl ProxySession for Session {
@@ -476,13 +532,14 @@ impl ProxySession for Session {
     //println!("TLS closing[{:?}] temp->front: {:?}, temp->back: {:?}", self.token, *self.temp.front_buf, *self.temp.back_buf);
     self.http_mut().map(|http| http.close());
     self.metrics.service_stop();
+    self.cancel_timeouts();
     if let Err(e) = self.front_socket().shutdown(Shutdown::Both) {
       if e.kind() != ErrorKind::NotConnected {
         error!("error closing front socket: {:?}", e);
       }
     }
 
-    if let Err(e) = poll.deregister(self.front_socket()) {
+    if let Err(e) = poll.registry().deregister(self.front_socket_mut()) {
       error!("error deregistering front socket: {:?}", e);
     }
 
@@ -495,10 +552,21 @@ impl ProxySession for Session {
     //FIXME: should we really pass a token here?
     self.close_backend(Token(0), poll);
 
-    if let Some(State::Http(ref http)) = self.protocol {
+    if let Some(State::Http(ref mut http)) = self.protocol {
       //if the state was initial, the connection was already reset
       if http.request != Some(RequestState::Initial) {
         gauge_add!("http.active_requests", -1);
+        if let Some(b) = http.backend_data.as_mut() {
+          let mut backend = b.borrow_mut();
+          backend.active_requests = backend.active_requests.saturating_sub(1);
+        }
+      }
+    }
+
+    if let Some(State::WebSocket(_)) = self.protocol {
+      if let Some(b) = self.backend.as_mut() {
+        let mut backend = b.borrow_mut();
+        backend.active_requests = backend.active_requests.saturating_sub(1);
       }
     }
 
@@ -515,37 +583,18 @@ impl ProxySession for Session {
     result
   }
 
-  fn timeout(&mut self, token: Token, timer: &mut Timer<Token>, front_timeout: &Duration) -> SessionResult {
-    if self.frontend_token == token {
-      let dur = Instant::now() - self.last_event;
-      if dur < *front_timeout {
-        timer.set_timeout(std::time::Duration::try_from(*front_timeout - dur).unwrap(), token);
-        SessionResult::Continue
-      } else {
-        match self.http().map(|h| h.timeout_status()) {
-          Some(TimeoutStatus::Request) => {
-            let answer = self.answers.borrow().get(DefaultAnswerStatus::Answer408, None);
-            self.set_answer(DefaultAnswerStatus::Answer408, answer);
-            self.writable()
-          },
-          Some(TimeoutStatus::Response) => {
-            let answer = self.answers.borrow().get(DefaultAnswerStatus::Answer504, None);
-            self.set_answer(DefaultAnswerStatus::Answer504, answer);
-            self.writable()
-          },
-          _ => {
-            SessionResult::CloseSession
+  fn timeout(&mut self, token: Token) -> SessionResult {
+    match *unwrap_msg!(self.protocol.as_mut()) {
+      State::Http(ref mut http) => http.timeout(token, &mut self.metrics),
+      State::WebSocket(ref mut pipe) => pipe.timeout(token, &mut self.metrics),
+      _ => {
+          if token == self.frontend_token {
+              self.front_timeout.triggered();
           }
-        }
-      }
-    } else {
-      //invalid token, obsolete timeout triggered
-      SessionResult::Continue
-    }
-  }
 
-  fn cancel_timeouts(&self, timer: &mut Timer<Token>) {
-    timer.cancel_timeout(&self.timeout);
+          SessionResult::CloseSession
+      },
+    }
   }
 
   fn close_backend(&mut self, _: Token, poll: &mut Poll) {
@@ -553,15 +602,15 @@ impl ProxySession for Session {
 
     let back_connected = self.back_connected();
     if back_connected != BackendConnectionStatus::NotConnected {
-      self.back_readiness().map(|r| r.event = UnixReady::from(Ready::empty()));
-      if let Some(sock) = self.back_socket() {
+      self.back_readiness().map(|r| r.event = Ready::empty());
+      if let Some(sock) = self.back_socket_mut() {
         if let Err(e) = sock.shutdown(Shutdown::Both) {
           if e.kind() != ErrorKind::NotConnected {
             error!("error shutting down backend socket: {:?}", e);
           }
         }
 
-        if let Err(e) = poll.deregister(sock) {
+        if let Err(e) = poll.registry().deregister(sock) {
           error!("error deregistering backend socket: {:?}", e);
         }
       }
@@ -583,14 +632,14 @@ impl ProxySession for Session {
   }
 
   fn process_events(&mut self, token: Token, events: Ready) {
-    trace!("token {:?} got event {}", token, super::super::unix_ready_to_string(UnixReady::from(events)));
+    trace!("token {:?} got event {}", token, super::super::ready_to_string(Ready::from(events)));
     self.last_event = Instant::now();
     self.metrics.wait_start();
 
     if self.frontend_token == token {
-      self.front_readiness().event = self.front_readiness().event | UnixReady::from(events);
+      self.front_readiness().event = self.front_readiness().event | Ready::from(events);
     } else if self.back_token() == Some(token) {
-      self.back_readiness().map(|r| r.event = r.event | UnixReady::from(events));
+      self.back_readiness().map(|r| r.event = r.event | Ready::from(events));
     }
   }
 
@@ -600,8 +649,10 @@ impl ProxySession for Session {
 
     self.metrics().service_start();
 
-    if self.back_connected() == BackendConnectionStatus::Connecting &&
-      self.back_readiness().map(|r| r.event != UnixReady::from(Ready::empty())).unwrap_or(false) {
+    if self.back_connected().is_connecting() &&
+      self.back_readiness().map(|r| r.event != Ready::empty()).unwrap_or(false) {
+
+      self.http_mut().map(|h| h.cancel_backend_timeout());
 
       if self.back_readiness().map(|r| r.event.is_hup()).unwrap_or(false) ||
         !self.http_mut().map(|h| h.test_back_socket()).unwrap_or(false) {
@@ -613,6 +664,7 @@ impl ProxySession for Session {
         self.fail_backend_connection();
 
         let backend_token = self.back_token();
+        self.back_connected = BackendConnectionStatus::Connecting(Instant::now());
         return SessionResult::ReconnectBackend(Some(self.frontend_token), backend_token);
       } else {
         self.metrics().backend_connected();
@@ -628,7 +680,7 @@ impl ProxySession for Session {
           return order;
         },
         _ => {
-          self.front_readiness().event.remove(UnixReady::hup());
+          self.front_readiness().event.remove(Ready::hup());
           return order;
         }
       }
@@ -637,11 +689,11 @@ impl ProxySession for Session {
     let token = self.frontend_token;
     while counter < max_loop_iterations {
       let front_interest = self.front_readiness().interest & self.front_readiness().event;
-      let back_interest  = self.back_readiness().map(|r| r.interest & r.event).unwrap_or(UnixReady::from(Ready::empty()));
+      let back_interest  = self.back_readiness().map(|r| r.interest & r.event).unwrap_or(Ready::empty());
 
       trace!("PROXY\t{} {:?} F:{:?} B:{:?}", self.log_context(), token, self.front_readiness().clone(), self.back_readiness());
 
-      if front_interest == UnixReady::from(Ready::empty()) && back_interest == UnixReady::from(Ready::empty()) {
+      if front_interest == Ready::empty() && back_interest == Ready::empty() {
         break;
       }
 
@@ -689,7 +741,7 @@ impl ProxySession for Session {
           },
           SessionResult::Continue => {},
           _ => {
-            self.back_readiness().map(|r| r.event.remove(UnixReady::hup()));
+            self.back_readiness().map(|r| r.event.remove(Ready::hup()));
             return order;
           }
         };
@@ -697,16 +749,15 @@ impl ProxySession for Session {
 
       if front_interest.is_error() {
           error!("PROXY session {:?} front error, disconnecting", self.frontend_token);
-
-          self.front_readiness().interest = UnixReady::from(Ready::empty());
-          self.back_readiness().map(|r| r.interest  = UnixReady::from(Ready::empty()));
-              return SessionResult::CloseSession;
+          self.front_readiness().interest = Ready::empty();
+          self.back_readiness().map(|r| r.interest  = Ready::empty());
+          return SessionResult::CloseSession;
       }
 
       if back_interest.is_error() {
           if self.back_hup() == SessionResult::CloseSession {
-              self.front_readiness().interest = UnixReady::from(Ready::empty());
-              self.back_readiness().map(|r| r.interest  = UnixReady::from(Ready::empty()));
+              self.front_readiness().interest = Ready::empty();
+              self.back_readiness().map(|r| r.interest  = Ready::empty());
               error!("PROXY session {:?} back error, disconnecting", self.frontend_token);
               return SessionResult::CloseSession;
           }
